@@ -1,30 +1,31 @@
 """pytest-asyncio implementation."""
 
+from __future__ import annotations
+
 import asyncio
 import contextlib
+import contextvars
 import enum
 import functools
 import inspect
 import socket
 import warnings
-from asyncio import AbstractEventLoopPolicy
-from textwrap import dedent
-from typing import (
-    Any,
+from asyncio import AbstractEventLoop, AbstractEventLoopPolicy
+from collections.abc import (
     AsyncIterator,
     Awaitable,
-    Callable,
-    Dict,
+    Coroutine as AbstractCoroutine,
     Generator,
     Iterable,
     Iterator,
-    List,
-    Literal,
     Mapping,
-    Optional,
     Sequence,
-    Set,
-    Type,
+)
+from textwrap import dedent
+from typing import (
+    Any,
+    Callable,
+    Literal,
     TypeVar,
     Union,
     overload,
@@ -112,16 +113,16 @@ def pytest_addoption(parser: Parser, pluginmanager: PytestPluginManager) -> None
 def fixture(
     fixture_function: FixtureFunction,
     *,
-    scope: "Union[_ScopeName, Callable[[str, Config], _ScopeName]]" = ...,
-    loop_scope: Union[_ScopeName, None] = ...,
-    params: Optional[Iterable[object]] = ...,
+    scope: _ScopeName | Callable[[str, Config], _ScopeName] = ...,
+    loop_scope: _ScopeName | None = ...,
+    params: Iterable[object] | None = ...,
     autouse: bool = ...,
-    ids: Union[
-        Iterable[Union[str, float, int, bool, None]],
-        Callable[[Any], Optional[object]],
-        None,
-    ] = ...,
-    name: Optional[str] = ...,
+    ids: (
+        Iterable[str | float | int | bool | None]
+        | Callable[[Any], object | None]
+        | None
+    ) = ...,
+    name: str | None = ...,
 ) -> FixtureFunction: ...
 
 
@@ -129,24 +130,24 @@ def fixture(
 def fixture(
     fixture_function: None = ...,
     *,
-    scope: "Union[_ScopeName, Callable[[str, Config], _ScopeName]]" = ...,
-    loop_scope: Union[_ScopeName, None] = ...,
-    params: Optional[Iterable[object]] = ...,
+    scope: _ScopeName | Callable[[str, Config], _ScopeName] = ...,
+    loop_scope: _ScopeName | None = ...,
+    params: Iterable[object] | None = ...,
     autouse: bool = ...,
-    ids: Union[
-        Iterable[Union[str, float, int, bool, None]],
-        Callable[[Any], Optional[object]],
-        None,
-    ] = ...,
-    name: Optional[str] = None,
+    ids: (
+        Iterable[str | float | int | bool | None]
+        | Callable[[Any], object | None]
+        | None
+    ) = ...,
+    name: str | None = None,
 ) -> FixtureFunctionMarker: ...
 
 
 def fixture(
-    fixture_function: Optional[FixtureFunction] = None,
-    loop_scope: Union[_ScopeName, None] = None,
+    fixture_function: FixtureFunction | None = None,
+    loop_scope: _ScopeName | None = None,
     **kwargs: Any,
-) -> Union[FixtureFunction, FixtureFunctionMarker]:
+) -> FixtureFunction | FixtureFunctionMarker:
     if fixture_function is not None:
         _make_asyncio_fixture_function(fixture_function, loop_scope)
         return pytest.fixture(fixture_function, **kwargs)
@@ -165,9 +166,7 @@ def _is_asyncio_fixture_function(obj: Any) -> bool:
     return getattr(obj, "_force_asyncio_fixture", False)
 
 
-def _make_asyncio_fixture_function(
-    obj: Any, loop_scope: Union[_ScopeName, None]
-) -> None:
+def _make_asyncio_fixture_function(obj: Any, loop_scope: _ScopeName | None) -> None:
     if hasattr(obj, "__func__"):
         # instance method, check the function object
         obj = obj.__func__
@@ -176,7 +175,7 @@ def _make_asyncio_fixture_function(
 
 
 def _is_coroutine_or_asyncgen(obj: Any) -> bool:
-    return asyncio.iscoroutinefunction(obj) or inspect.isasyncgenfunction(obj)
+    return inspect.iscoroutinefunction(obj) or inspect.isasyncgenfunction(obj)
 
 
 def _get_asyncio_mode(config: Config) -> Mode:
@@ -185,11 +184,11 @@ def _get_asyncio_mode(config: Config) -> Mode:
         val = config.getini("asyncio_mode")
     try:
         return Mode(val)
-    except ValueError:
+    except ValueError as e:
         modes = ", ".join(m.value for m in Mode)
         raise pytest.UsageError(
             f"{val!r} is not a valid asyncio_mode. Valid modes: {modes}."
-        )
+        ) from e
 
 
 _DEFAULT_FIXTURE_LOOP_SCOPE_UNSET = """\
@@ -215,16 +214,18 @@ def pytest_configure(config: Config) -> None:
 
 
 @pytest.hookimpl(tryfirst=True)
-def pytest_report_header(config: Config) -> List[str]:
+def pytest_report_header(config: Config) -> list[str]:
     """Add asyncio config to pytest header."""
     mode = _get_asyncio_mode(config)
     default_loop_scope = config.getini("asyncio_default_fixture_loop_scope")
-    return [f"asyncio: mode={mode}, default_loop_scope={default_loop_scope}"]
+    return [
+        f"asyncio: mode={mode}, asyncio_default_fixture_loop_scope={default_loop_scope}"
+    ]
 
 
 def _preprocess_async_fixtures(
     collector: Collector,
-    processed_fixturedefs: Set[FixtureDef],
+    processed_fixturedefs: set[FixtureDef],
 ) -> None:
     config = collector.config
     default_loop_scope = config.getini("asyncio_default_fixture_loop_scope")
@@ -238,7 +239,7 @@ def _preprocess_async_fixtures(
                 func
             ):
                 continue
-            if not _is_asyncio_fixture_function(func) and asyncio_mode == Mode.STRICT:
+            if asyncio_mode == Mode.STRICT and not _is_asyncio_fixture_function(func):
                 # Ignore async fixtures without explicit asyncio mark in strict mode
                 # This applies to pytest_trio fixtures, for example
                 continue
@@ -268,9 +269,7 @@ def _preprocess_async_fixtures(
 
 
 def _synchronize_async_fixture(fixturedef: FixtureDef) -> None:
-    """
-    Wraps the fixture function of an async fixture in a synchronous function.
-    """
+    """Wraps the fixture function of an async fixture in a synchronous function."""
     if inspect.isasyncgenfunction(fixturedef.func):
         _wrap_asyncgen_fixture(fixturedef)
     elif inspect.iscoroutinefunction(fixturedef.func):
@@ -279,10 +278,10 @@ def _synchronize_async_fixture(fixturedef: FixtureDef) -> None:
 
 def _add_kwargs(
     func: Callable[..., Any],
-    kwargs: Dict[str, Any],
+    kwargs: dict[str, Any],
     event_loop: asyncio.AbstractEventLoop,
     request: FixtureRequest,
-) -> Dict[str, Any]:
+) -> dict[str, Any]:
     sig = inspect.signature(func)
     ret = kwargs.copy()
     if "request" in sig.parameters:
@@ -292,7 +291,7 @@ def _add_kwargs(
     return ret
 
 
-def _perhaps_rebind_fixture_func(func: _T, instance: Optional[Any]) -> _T:
+def _perhaps_rebind_fixture_func(func: _T, instance: Any | None) -> _T:
     if instance is not None:
         # The fixture needs to be bound to the actual request.instance
         # so it is bound to the same object as the test method.
@@ -325,6 +324,12 @@ def _wrap_asyncgen_fixture(fixturedef: FixtureDef) -> None:
             res = await gen_obj.__anext__()  # type: ignore[union-attr]
             return res
 
+        context = contextvars.copy_context()
+        setup_task = _create_task_in_context(event_loop, setup(), context)
+        result = event_loop.run_until_complete(setup_task)
+
+        reset_contextvars = _apply_contextvar_changes(context)
+
         def finalizer() -> None:
             """Yield again, to finalize."""
 
@@ -338,9 +343,11 @@ def _wrap_asyncgen_fixture(fixturedef: FixtureDef) -> None:
                     msg += "Yield only once."
                     raise ValueError(msg)
 
-            event_loop.run_until_complete(async_finalizer())
+            task = _create_task_in_context(event_loop, async_finalizer(), context)
+            event_loop.run_until_complete(task)
+            if reset_contextvars is not None:
+                reset_contextvars()
 
-        result = event_loop.run_until_complete(setup())
         request.addfinalizer(finalizer)
         return result
 
@@ -363,7 +370,23 @@ def _wrap_async_fixture(fixturedef: FixtureDef) -> None:
             res = await func(**_add_kwargs(func, kwargs, event_loop, request))
             return res
 
-        return event_loop.run_until_complete(setup())
+        context = contextvars.copy_context()
+        setup_task = _create_task_in_context(event_loop, setup(), context)
+        result = event_loop.run_until_complete(setup_task)
+
+        # Copy the context vars modified by the setup task into the current
+        # context, and (if needed) add a finalizer to reset them.
+        #
+        # Note that this is slightly different from the behavior of a non-async
+        # fixture, which would rely on the fixture author to add a finalizer
+        # to reset the variables. In this case, the author of the fixture can't
+        # write such a finalizer because they have no way to capture the Context
+        # in which the setup function was run, so we need to do it for them.
+        reset_contextvars = _apply_contextvar_changes(context)
+        if reset_contextvars is not None:
+            request.addfinalizer(reset_contextvars)
+
+        return result
 
     fixturedef.func = _async_fixture_wrapper  # type: ignore[misc]
 
@@ -388,13 +411,66 @@ def _get_event_loop_fixture_id_for_async_fixture(
     return event_loop_fixture_id
 
 
+def _create_task_in_context(
+    loop: asyncio.AbstractEventLoop,
+    coro: AbstractCoroutine[Any, Any, _T],
+    context: contextvars.Context,
+) -> asyncio.Task[_T]:
+    """
+    Return an asyncio task that runs the coro in the specified context,
+    if possible.
+
+    This allows fixture setup and teardown to be run as separate asyncio tasks,
+    while still being able to use context-manager idioms to maintain context
+    variables and make those variables visible to test functions.
+
+    This is only fully supported on Python 3.11 and newer, as it requires
+    the API added for https://github.com/python/cpython/issues/91150.
+    On earlier versions, the returned task will use the default context instead.
+    """
+    try:
+        return loop.create_task(coro, context=context)
+    except TypeError:
+        return loop.create_task(coro)
+
+
+def _apply_contextvar_changes(
+    context: contextvars.Context,
+) -> Callable[[], None] | None:
+    """
+    Copy contextvar changes from the given context to the current context.
+
+    If any contextvars were modified by the fixture, return a finalizer that
+    will restore them.
+    """
+    context_tokens = []
+    for var in context:
+        try:
+            if var.get() is context.get(var):
+                # This variable is not modified, so leave it as-is.
+                continue
+        except LookupError:
+            # This variable isn't yet set in the current context at all.
+            pass
+        token = var.set(context.get(var))
+        context_tokens.append((var, token))
+
+    if not context_tokens:
+        return None
+
+    def restore_contextvars():
+        while context_tokens:
+            (var, token) = context_tokens.pop()
+            var.reset(token)
+
+    return restore_contextvars
+
+
 class PytestAsyncioFunction(Function):
     """Base class for all test functions managed by pytest-asyncio."""
 
     @classmethod
-    def item_subclass_for(
-        cls, item: Function, /
-    ) -> Union[Type["PytestAsyncioFunction"], None]:
+    def item_subclass_for(cls, item: Function, /) -> type[PytestAsyncioFunction] | None:
         """
         Returns a subclass of PytestAsyncioFunction if there is a specialized subclass
         for the specified function item.
@@ -447,7 +523,7 @@ class Coroutine(PytestAsyncioFunction):
     @staticmethod
     def _can_substitute(item: Function) -> bool:
         func = item.obj
-        return asyncio.iscoroutinefunction(func)
+        return inspect.iscoroutinefunction(func)
 
     def runtest(self) -> None:
         self.obj = wrap_in_sync(
@@ -512,7 +588,7 @@ class AsyncHypothesisTest(PytestAsyncioFunction):
         return (
             getattr(func, "is_hypothesis_test", False)  # type: ignore[return-value]
             and getattr(func, "hypothesis", None)
-            and asyncio.iscoroutinefunction(func.hypothesis.inner_test)
+            and inspect.iscoroutinefunction(func.hypothesis.inner_test)
         )
 
     def runtest(self) -> None:
@@ -522,17 +598,15 @@ class AsyncHypothesisTest(PytestAsyncioFunction):
         super().runtest()
 
 
-_HOLDER: Set[FixtureDef] = set()
+_HOLDER: set[FixtureDef] = set()
 
 
 # The function name needs to start with "pytest_"
 # see https://github.com/pytest-dev/pytest/issues/11307
 @pytest.hookimpl(specname="pytest_pycollect_makeitem", tryfirst=True)
 def pytest_pycollect_makeitem_preprocess_async_fixtures(
-    collector: Union[pytest.Module, pytest.Class], name: str, obj: object
-) -> Union[
-    pytest.Item, pytest.Collector, List[Union[pytest.Item, pytest.Collector]], None
-]:
+    collector: pytest.Module | pytest.Class, name: str, obj: object
+) -> pytest.Item | pytest.Collector | list[pytest.Item | pytest.Collector] | None:
     """A pytest hook to collect asyncio coroutines."""
     if not collector.funcnamefilter(name):
         return None
@@ -544,7 +618,7 @@ def pytest_pycollect_makeitem_preprocess_async_fixtures(
 # see https://github.com/pytest-dev/pytest/issues/11307
 @pytest.hookimpl(specname="pytest_pycollect_makeitem", hookwrapper=True)
 def pytest_pycollect_makeitem_convert_async_functions_to_subclass(
-    collector: Union[pytest.Module, pytest.Class], name: str, obj: object
+    collector: pytest.Module | pytest.Class, name: str, obj: object
 ) -> Generator[None, pluggy.Result, None]:
     """
     Converts coroutines and async generators collected as pytest.Functions
@@ -552,12 +626,9 @@ def pytest_pycollect_makeitem_convert_async_functions_to_subclass(
     """
     hook_result = yield
     try:
-        node_or_list_of_nodes: Union[
-            pytest.Item,
-            pytest.Collector,
-            List[Union[pytest.Item, pytest.Collector]],
-            None,
-        ] = hook_result.get_result()
+        node_or_list_of_nodes: (
+            pytest.Item | pytest.Collector | list[pytest.Item | pytest.Collector] | None
+        ) = hook_result.get_result()
     except BaseException as e:
         hook_result.force_exception(e)
         return
@@ -585,7 +656,7 @@ def pytest_pycollect_makeitem_convert_async_functions_to_subclass(
 
 
 _event_loop_fixture_id = StashKey[str]()
-_fixture_scope_by_collector_type: Mapping[Type[pytest.Collector], _ScopeName] = {
+_fixture_scope_by_collector_type: Mapping[type[pytest.Collector], _ScopeName] = {
     Class: "class",
     # Package is a subclass of module and the dict is used in isinstance checks
     # Therefore, the order matters and Package needs to appear before Module
@@ -596,7 +667,7 @@ _fixture_scope_by_collector_type: Mapping[Type[pytest.Collector], _ScopeName] =
 
 # A stack used to push package-scoped loops during collection of a package
 # and pop those loops during collection of a Module
-__package_loop_stack: List[Union[FixtureFunctionMarker, FixtureFunction]] = []
+__package_loop_stack: list[FixtureFunctionMarker | FixtureFunction] = []
 
 
 @pytest.hookimpl
@@ -637,12 +708,12 @@ def pytest_collectstart(collector: pytest.Collector) -> None:
         event_loop_policy,
     ) -> Iterator[asyncio.AbstractEventLoop]:
         new_loop_policy = event_loop_policy
-        with _temporary_event_loop_policy(new_loop_policy):
-            loop = asyncio.new_event_loop()
-            loop.__pytest_asyncio = True  # type: ignore[attr-defined]
+        with (
+            _temporary_event_loop_policy(new_loop_policy),
+            _provide_event_loop() as loop,
+        ):
             asyncio.set_event_loop(loop)
             yield loop
-            loop.close()
 
     # @pytest.fixture does not register the fixture anywhere, so pytest doesn't
     # know it exists. We work around this by attaching the fixture function to the
@@ -691,6 +762,19 @@ def _temporary_event_loop_policy(policy: AbstractEventLoopPolicy) -> Iterator[No
     try:
         yield
     finally:
+        # Try detecting user-created event loops that were left unclosed
+        # at the end of a test.
+        try:
+            current_loop: AbstractEventLoop | None = _get_event_loop_no_warn()
+        except RuntimeError:
+            current_loop = None
+        if current_loop is not None and not current_loop.is_closed():
+            warnings.warn(
+                _UNCLOSED_EVENT_LOOP_WARNING % current_loop,
+                DeprecationWarning,
+            )
+            current_loop.close()
+
         asyncio.set_event_loop_policy(old_loop_policy)
         # When a test uses both a scoped event loop and the event_loop fixture,
         # the "_provide_clean_event_loop" finalizer of the event_loop fixture
@@ -711,7 +795,7 @@ _REDEFINED_EVENT_LOOP_FIXTURE_WARNING = dedent(
     Replacing the event_loop fixture with a custom implementation is deprecated
     and will lead to errors in the future.
     If you want to request an asyncio event loop with a scope other than function
-    scope, use the "scope" argument to the asyncio mark when marking the tests.
+    scope, use the "loop_scope" argument to the asyncio mark when marking the tests.
     If you want to return different types of event loops, use the event_loop_policy
     fixture.
     """
@@ -778,7 +862,7 @@ def pytest_fixture_setup(
         # Weird behavior was observed when checking for an attribute of FixtureDef.func
         # Instead, we now check for a special attribute of the returned event loop
         fixture_filename = inspect.getsourcefile(fixturedef.func)
-        if not getattr(loop, "__original_fixture_loop", False):
+        if not _is_pytest_asyncio_loop(loop):
             _, fixture_line_number = inspect.getsourcelines(fixturedef.func)
             warnings.warn(
                 _REDEFINED_EVENT_LOOP_FIXTURE_WARNING
@@ -788,8 +872,7 @@ def pytest_fixture_setup(
         policy = asyncio.get_event_loop_policy()
         try:
             old_loop = _get_event_loop_no_warn(policy)
-            is_pytest_asyncio_loop = getattr(old_loop, "__pytest_asyncio", False)
-            if old_loop is not loop and not is_pytest_asyncio_loop:
+            if old_loop is not loop and not _is_pytest_asyncio_loop(old_loop):
                 old_loop.close()
         except RuntimeError:
             # Either the current event loop has been set to None
@@ -802,11 +885,20 @@ def pytest_fixture_setup(
     yield
 
 
+def _make_pytest_asyncio_loop(loop: AbstractEventLoop) -> AbstractEventLoop:
+    loop.__pytest_asyncio = True  # type: ignore[attr-defined]
+    return loop
+
+
+def _is_pytest_asyncio_loop(loop: AbstractEventLoop) -> bool:
+    return getattr(loop, "__pytest_asyncio", False)
+
+
 def _add_finalizers(fixturedef: FixtureDef, *finalizers: Callable[[], object]) -> None:
     """
-    Regsiters the specified fixture finalizers in the fixture.
+    Registers the specified fixture finalizers in the fixture.
 
-    Finalizers need to specified in the exact order in which they should be invoked.
+    Finalizers need to be specified in the exact order in which they should be invoked.
 
     :param fixturedef: Fixture definition which finalizers should be added to
     :param finalizers: Finalizers to be added
@@ -835,7 +927,7 @@ def _close_event_loop() -> None:
         loop = policy.get_event_loop()
     except RuntimeError:
         loop = None
-    if loop is not None:
+    if loop is not None and not _is_pytest_asyncio_loop(loop):
         if not loop.is_closed():
             warnings.warn(
                 _UNCLOSED_EVENT_LOOP_WARNING % loop,
@@ -852,7 +944,7 @@ def _restore_event_loop_policy(previous_policy) -> Callable[[], None]:
             loop = _get_event_loop_no_warn(previous_policy)
         except RuntimeError:
             loop = None
-        if loop:
+        if loop and not _is_pytest_asyncio_loop(loop):
             loop.close()
         asyncio.set_event_loop_policy(previous_policy)
 
@@ -867,12 +959,17 @@ def _provide_clean_event_loop() -> None:
     # Note that we cannot set the loop to None, because get_event_loop only creates
     # a new loop, when set_event_loop has not been called.
     policy = asyncio.get_event_loop_policy()
-    new_loop = policy.new_event_loop()
-    policy.set_event_loop(new_loop)
+    try:
+        old_loop = _get_event_loop_no_warn(policy)
+    except RuntimeError:
+        old_loop = None
+    if old_loop is not None and not _is_pytest_asyncio_loop(old_loop):
+        new_loop = policy.new_event_loop()
+        policy.set_event_loop(new_loop)
 
 
 def _get_event_loop_no_warn(
-    policy: Optional[AbstractEventLoopPolicy] = None,
+    policy: AbstractEventLoopPolicy | None = None,
 ) -> asyncio.AbstractEventLoop:
     with warnings.catch_warnings():
         warnings.simplefilter("ignore", DeprecationWarning)
@@ -883,7 +980,7 @@ def _get_event_loop_no_warn(
 
 
 @pytest.hookimpl(tryfirst=True, hookwrapper=True)
-def pytest_pyfunc_call(pyfuncitem: Function) -> Optional[object]:
+def pytest_pyfunc_call(pyfuncitem: Function) -> object | None:
     """
     Pytest hook called before a test case is run.
 
@@ -892,7 +989,32 @@ def pytest_pyfunc_call(pyfuncitem: Function) -> Optional[object]:
     """
     if pyfuncitem.get_closest_marker("asyncio") is not None:
         if isinstance(pyfuncitem, PytestAsyncioFunction):
-            pass
+            asyncio_mode = _get_asyncio_mode(pyfuncitem.config)
+            for fixname, fixtures in pyfuncitem._fixtureinfo.name2fixturedefs.items():
+                # name2fixturedefs is a dict between fixture name and a list of matching
+                # fixturedefs. The last entry in the list is closest and the one used.
+                func = fixtures[-1].func
+                if (
+                    asyncio_mode == Mode.STRICT
+                    and _is_coroutine_or_asyncgen(func)
+                    and not _is_asyncio_fixture_function(func)
+                ):
+                    warnings.warn(
+                        PytestDeprecationWarning(
+                            f"asyncio test {pyfuncitem.name!r} requested async "
+                            "@pytest.fixture "
+                            f"{fixname!r} in strict mode. "
+                            "You might want to use @pytest_asyncio.fixture or switch "
+                            "to auto mode. "
+                            "This will become an error in future versions of "
+                            "flake8-asyncio."
+                        ),
+                        stacklevel=1,
+                    )
+                    # no stacklevel points at the users code, so we set stacklevel=1
+                    # so it at least indicates that it's the plugin complaining.
+                    # Pytest gives the test file & name in the warnings summary at least
+
         else:
             pyfuncitem.warn(
                 pytest.PytestWarning(
@@ -910,9 +1032,10 @@ def pytest_pyfunc_call(pyfuncitem: Function) -> Optional[object]:
 def wrap_in_sync(
     func: Callable[..., Awaitable[Any]],
 ):
-    """Return a sync wrapper around an async function executing it in the
-    current event loop."""
-
+    """
+    Return a sync wrapper around an async function executing it in the
+    current event loop.
+    """
     # if the function is already wrapped, we rewrap using the original one
     # not using __wrapped__ because the original function may already be
     # a wrapped one
@@ -990,7 +1113,7 @@ def _get_marked_loop_scope(asyncio_marker: Mark) -> _ScopeName:
     if asyncio_marker.args or (
         asyncio_marker.kwargs and set(asyncio_marker.kwargs) - {"loop_scope", "scope"}
     ):
-        raise ValueError("mark.asyncio accepts only a keyword argument 'scope'.")
+        raise ValueError("mark.asyncio accepts only a keyword argument 'loop_scope'.")
     if "scope" in asyncio_marker.kwargs:
         if "loop_scope" in asyncio_marker.kwargs:
             raise pytest.UsageError(_DUPLICATE_LOOP_SCOPE_DEFINITION_ERROR)
@@ -1002,7 +1125,7 @@ def _get_marked_loop_scope(asyncio_marker: Mark) -> _ScopeName:
     return scope
 
 
-def _retrieve_scope_root(item: Union[Collector, Item], scope: str) -> Collector:
+def _retrieve_scope_root(item: Collector | Item, scope: str) -> Collector:
     node_type_by_scope = {
         "class": Class,
         "module": Module,
@@ -1025,16 +1148,26 @@ def _retrieve_scope_root(item: Union[Collector, Item], scope: str) -> Collector:
 def event_loop(request: FixtureRequest) -> Iterator[asyncio.AbstractEventLoop]:
     """Create an instance of the default event loop for each test case."""
     new_loop_policy = request.getfixturevalue(event_loop_policy.__name__)
-    asyncio.set_event_loop_policy(new_loop_policy)
+    with _temporary_event_loop_policy(new_loop_policy), _provide_event_loop() as loop:
+        yield loop
+
+
+@contextlib.contextmanager
+def _provide_event_loop() -> Iterator[asyncio.AbstractEventLoop]:
     loop = asyncio.get_event_loop_policy().new_event_loop()
     # Add a magic value to the event loop, so pytest-asyncio can determine if the
     # event_loop fixture was overridden. Other implementations of event_loop don't
     # set this value.
     # The magic value must be set as part of the function definition, because pytest
     # seems to have multiple instances of the same FixtureDef or fixture function
-    loop.__original_fixture_loop = True  # type: ignore[attr-defined]
-    yield loop
-    loop.close()
+    loop = _make_pytest_asyncio_loop(loop)
+    try:
+        yield loop
+    finally:
+        try:
+            loop.run_until_complete(loop.shutdown_asyncgens())
+        finally:
+            loop.close()
 
 
 @pytest.fixture(scope="session")
@@ -1042,12 +1175,9 @@ def _session_event_loop(
     request: FixtureRequest, event_loop_policy: AbstractEventLoopPolicy
 ) -> Iterator[asyncio.AbstractEventLoop]:
     new_loop_policy = event_loop_policy
-    with _temporary_event_loop_policy(new_loop_policy):
-        loop = asyncio.new_event_loop()
-        loop.__pytest_asyncio = True  # type: ignore[attr-defined]
+    with _temporary_event_loop_policy(new_loop_policy), _provide_event_loop() as loop:
         asyncio.set_event_loop(loop)
         yield loop
-        loop.close()
 
 
 @pytest.fixture(scope="session", autouse=True)
